{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE DeriveDataTypeable #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE StandaloneDeriving #-}
{-# LANGUAGE DerivingStrategies #-}

module Language.Haskell.Brittany.Internal.Types where

import qualified Control.Monad.Trans.MultiRWS.Strict as MultiRWSS
import qualified Data.Data
import qualified Data.Kind as Kind
import qualified GHC.OldList as List
import qualified Data.Text.Lazy.Builder as Text.Builder
import           GHC                            ( Anno
                                                , DeltaPos
                                                  ( DifferentLine
                                                  , SameLine
                                                  )
                                                , EpaCommentTok
                                                , LHsDecl
                                                , ParsedSource
                                                , XRec
                                                , LImportDecl
                                                )
import GHC.Utils.Outputable(Outputable)
import Language.Haskell.Brittany.Internal.Config.Types
import Language.Haskell.Brittany.Internal.Prelude
import Language.Haskell.Brittany.Internal.Components.BriDoc



-- brittany-internal error type, part of the brittany library public interface.
data BrittanyError
  = ErrorInput String
    -- ^ parsing failed
  | ErrorUnusedComment String
    -- ^ internal error: some comment went missing. Since ghc-9.2 the one below
    --   is used for missing comments while layouting top-level decls.
    --   You might still get thse for import/export statement comments (right?)
    --   (TODO this needs updating once the ghc-9.2 implementation is complete)
  | ErrorUnusedComments (LHsDecl GhcPs) Int Int
    -- ^ internal error: some comments went missing while layouting the
    --   specified top-level declaration (module child).
  | ErrorMacroConfig String String
    -- ^ in-source config string parsing error; first argument is the parser
    --   output and second the corresponding, ill-formed input.
  | LayoutWarning String
    -- ^ some warning
  | forall ast . (Data.Data.Data (XRec GhcPs ast), Outputable (Anno ast)) => ErrorUnknownNode String (XRec GhcPs ast)
    -- ^ internal error: pretty-printing is not implemented for type of node
    --   in the syntax-tree
  | ErrorOutputCheck
    -- ^ checking the output for syntactic validity failed


-- General-purpose structure. Church-encoded version of a list that has a
-- parameterized final value.
-- That is:
-- 1) Start with a dumb old `List a = Cons a (List a) | Empty`
-- 2) Parameterize the empty `ListEnd a b = Cons a (List a b) | Final b`
-- 3) Church-encode the result for good measure (and better perf?)
--
-- This is similar to a Writer Monad, and maybe it already exists somewhere,
-- but it feels like a) This abstraction doesn't leak by itself b) you can
-- trivially turn it into something that streams into some "consuming" monadic
-- action c) it is _simple_, doesn't require a monoid or anything without
-- losing anything.
data FinalList a b = FinalList (forall r . (a -> r -> r) -> (b -> r) -> r)

finalPure :: b -> FinalList a b
finalPure b = FinalList $ \_ f -> f b
finalCons :: a -> FinalList a b -> FinalList a b
finalCons a (FinalList l) = FinalList (\f1 f2 -> f1 a $ l f1 f2)
finalBind :: FinalList a b -> (b -> FinalList a b) -> FinalList a b
finalBind (FinalList l) f =
  FinalList (\f1 f2 -> l f1 (\b -> let FinalList c = (f b) in c f1 f2))
_finalSingleton :: a -> b -> FinalList a b
_finalSingleton a b = FinalList (\f1 f2 -> f1 a (f2 b))
_finalLMap :: (a -> a') -> FinalList a b -> FinalList a' b
_finalLMap f (FinalList l) = FinalList (\f1 f2 -> l (f1 . f) f2)
_finalRMap :: (b -> b') -> FinalList a b -> FinalList a b'
_finalRMap f (FinalList l) = FinalList (\f1 f2 -> l f1 (f2 . f))
finalYield :: a -> FinalList a ()
finalYield x = FinalList $ \f1 f2 -> f1 x (f2 ())
finalToList_ :: FinalList a () -> [a]
finalToList_ (FinalList l) = l (:) (\() -> [])
finalToList :: FinalList a b -> ([a], b)
finalToList (FinalList l) = l (\x (a, b) -> (x:a, b)) (\b -> ([], b))

instance Functor (FinalList a) where
  fmap = _finalRMap

instance Applicative (FinalList a) where
  pure = finalPure
  FinalList ff <*> FinalList fx =
    FinalList $ \f1 f2 -> ff f1 (\g -> fx f1 (f2 . g))

instance Monad (FinalList a) where
  FinalList fx >>= f =
    FinalList $ \f1 f2 -> fx f1 (\x -> let FinalList c = f x in c f1 f2)

-- After parsing, we split a module into a series of these so that
-- we can process them independently.
data ModuleElement
  = MEExactModuleHead ParsedSource
    -- ^ When not pretty-printing the module head:
    --   module name, imports, exports, but without decls.
    --   Could split out the relevant fields, but we need the full (with decls
    --   set to []) for exactprinting anyway.
  | MEPrettyModuleHead ParsedSource
    -- ^ Similar, but when pretty-printing the head:
    --   This encompasses just the module name, the exports and
    --   the where keyword. Everything else (including comments in
    --   various places) get 
  | MEImportDecl (LImportDecl GhcPs) [(Int, EpaCommentTok)]
    -- ^ an import decl, only occurs if pretty-printing the module head.
  | MEDecl (LHsDecl GhcPs) [(Int, EpaCommentTok)]
    -- ^ a top-level declaration
  | MEComment (Int, EpaCommentTok)
    -- ^ a top-level comment, i.e. a comment located between top-level elements
    --   (and not associated to some nested node, which might in theory happen).
    --   The Int carries the indentation of the comment.
  | MEWhitespace DeltaPos
  -- ^ Empty lines around decls.


newtype TraceFunc = TraceFunc { useTraceFunc :: String -> IO () }

type PPMLocal = MultiRWSS.MultiRWS
  '[Config, TraceFunc]
  '[Text.Builder.Builder, [BrittanyError], Seq String]
  '[CommentCounter]

-- type LayoutConstraints m
--   = ( MonadMultiReader Config m
--     , MonadMultiWriter TextL.Builder.Builder m
--     , MonadMultiWriter (Seq String) m
--     , MonadMultiState LayoutState m
--     , MonadMultiState CommentCounter m
--     , MonadMultiState [GHC.LEpaComment] m
--     )

ppmMoveToExactLoc
  :: MonadMultiWriter Text.Builder.Builder m => GHC.DeltaPos -> m ()
ppmMoveToExactLoc = \case
  SameLine c ->
    mTell $ Text.Builder.fromString (List.replicate c ' ')
  DifferentLine l c -> mTell $ Text.Builder.fromString
    (List.replicate l '\n' ++ List.replicate (c - 1) ' ')


type ToBriDocM = MultiRWSS.MultiRWS
  '[Config, TraceFunc] -- reader
  '[[BrittanyError], Seq String] -- writer
  '[NodeAllocIndex, CommentCounter] -- state

type ToBriDoc (sym :: Kind.Type -> Kind.Type) = XRec GhcPs (sym GhcPs) -> ToBriDocM BriDocNumbered
type ToBriDoc' sym            = XRec GhcPs (sym GhcPs)                 -> ToBriDocM BriDocNumbered
type ToBriDocC sym c          = XRec GhcPs (sym GhcPs)                 -> ToBriDocM c
type ToBriDocP sym            = sym GhcPs                              -> ToBriDocM BriDocNumbered


newtype CommentCounter = CommentCounter { unCommentCounter :: Int }
  deriving (Eq, Ord, Num, Data.Data.Data)
  deriving newtype (Show)

-- Why does this live in types? As long as it does, you _have_
-- to remember to call `docFlushRemaining` in combination with this!
briDocMToPPM :: ToBriDocM a -> PPMLocal (a, Int)
briDocMToPPM m = do
  readers <- MultiRWSS.mGetRawR
  initCount <- MultiRWSS.mGet @CommentCounter
  let
    (((x, errs), debugs), commentCount) =
      runIdentity
        $ MultiRWSS.runMultiRWSTNil
        $ MultiRWSS.withMultiStateAS initCount
        $ MultiRWSS.withMultiStateA (NodeAllocIndex 1)
        $ MultiRWSS.withMultiReaders readers
        $ MultiRWSS.withMultiWriterAW
        $ MultiRWSS.withMultiWriterAW
        $ m
  mTell debugs
  mTell errs
  mSet commentCount
  pure (x, length errs)



newtype NodeAllocIndex = NodeAllocIndex Int



-- data NewLineState = NewLineStateInit -- initial state. we do not know if in a
--                                      -- newline, really. by special-casing
--                                      -- this we can appropriately handle it
--                                      -- differently at use-site.
--                   | NewLineStateYes
--                   | NewLineStateNo
--   deriving Eq

-- data LayoutSettings = LayoutSettings
--   { _lsettings_cols :: Int -- the thing that has default 80.
--   , _lsettings_indentPolicy :: IndentPolicy
--   , _lsettings_indentAmount :: Int
--   , _lsettings_indentWhereSpecial :: Bool -- indent where only 1 sometimes (TODO).
--   , _lsettings_indentListSpecial  :: Bool -- use some special indentation for ","
--                                           -- when creating zero-indentation
--                                           -- multi-line list literals.
--   , _lsettings_importColumn :: Int
--   , _lsettings_initialAnns :: ExactPrint.Anns
--   }


-- TODO92 is this old leftover, or useful future idea?
-- data BriSpacing = BriSpacing
--   { _bs_spacePastLineIndent :: Int -- space in the current,
--                                    -- potentially somewhat filled
--                                    -- line.
--   , _bs_spacePastIndent :: Int     -- space required in properly
--                                    -- indented blocks below the
--                                    -- current line.
--   }
